@@ -194,6 +194,10 @@ static void ext2fs_copy_inode(struct ext2_inode * dino, FS_INODE *fs_inode)
     fs_inode->mode = dino->i_mode;
     fs_inode->nlink = dino->i_links_count;
     fs_inode->size = dino->i_size;
+#ifdef i_size_high
+    if (dino->i_size_high)
+	fs_inode->size |= (dino->i_size_high << 32);
+#endif
     fs_inode->uid = dino->i_uid;
     fs_inode->gid = dino->i_gid;
     fs_inode->mtime = dino->i_mtime;
