@@ -113,6 +113,24 @@ extern int optind;
 #define UFS_TYPE	"ufs"
 #define DEF_FSTYPE	UFS_TYPE
 #define INO_TO_CG	ino_to_cg
+#endif
+
+#if defined(FREEBSD5)
+#define SUPPORTED
+#include <sys/vnode.h>
+#include <ufs/ufs/quota.h>
+#include <ufs/ufs/inode.h>
+#include <ufs/ffs/fs.h>
+#define LSEEK		lseek
+#define OFF_T		off_t
+#define STRTOUL		strtoul
+#define itod(fs,i)	ino_to_fsba(fs,i)
+#define itoo(fs,i)	ino_to_fsbo(fs,i)
+#define INOTIME(t)	(t)
+#define DADDR_T		int64_t
+#define UFS_TYPE	"ufs"
+#define DEF_FSTYPE	UFS_TYPE
+#define INO_TO_CG	ino_to_cg
 #endif
 
  /*
